/*
	File: blop_c.c
	Author: Shivkumar Chandrasekaran
	Date: 6.17.2003
	Email: shiv@ece.ucsb.edu
	Purpose: C glue between Clean/OCaml and Lapack and Blas
	Package: CleanFloat and CamlFloat
	
	$Id: blop_c.c,v 1.1 2003/08/21 21:25:35 shiv Exp shiv $
*/

/* #include "/sw/include/cblas.h" */
#include <vecLib/cblas.h>

typedef struct {double re; double im;} complex;
typedef struct {float re; float im;} scomplex;

double dlange_ (char *norm, long int *m, long int *n,
	double *a, long int *lda, double *work);
	
double my_dlange (int norm, int m, int n, double *a, int disp, int lda, double *work)
{
	long int mm = (long int) m, nn = (long int) n, ldaa = (long int) lda;
	char nrm[] = {'O','F','I'};
	return (dlange_ (nrm+norm, &mm, &nn, a+disp, &ldaa, work));
}

double zlange_ (char *norm, long int *m, long int *n,
	complex *a, long int *lda, double *work);

double my_zlange (int norm, int m, int n, complex *a, int disp, int lda, double *work)
{
	long int mm = (long int) m, nn = (long int) n, ldaa = (long int) lda;
	char nrm[] = {'O','F','I'};
	return (zlange_ (nrm+norm, &mm, &nn, a+disp, &ldaa, work));
}

float slange_ (char *norm, long int *m, long int *n,
	float *a, long int *lda, float *work);
	
float my_slange (int norm, int m, int n, float *a, int disp, int lda, float *work)
{
	long int mm = (long int) m, nn = (long int) n, ldaa = (long int) lda;
	char nrm[] = {'O','F','I'};
	return (slange_ (nrm+norm, &mm, &nn, a+disp, &ldaa, work));
}

float clange_ (char *norm, long int *m, long int *n,
	scomplex *a, long int *lda, float *work);

double my_clange (int norm, int m, int n, scomplex *a, int disp, int lda, float *work)
{
	long int mm = (long int) m, nn = (long int) n, ldaa = (long int) lda;
	char nrm[] = {'O','F','I'};
	return (clange_ (nrm+norm, &mm, &nn, a+disp, &ldaa, work));
}



int my_cblas_dgemm (int order, int transA, int transB, int m, int n, int k, double alpha,
	double *a, int dispa, int lda, double *b, int dispb, int ldb, double beta,
	double *c, int dispc, int ldc) {
	
	cblas_dgemm (order, transA, transB, m, n, k, alpha, a+dispa, lda, b+dispb, ldb,
		beta, c+dispc, ldc);
	return 0;
}

int my_cblas_zgemm (int order, int transA, int transB, int m, int n, int k, complex *alpha,
	complex *a, int dispa, int lda, complex *b, int dispb, int ldb, complex *beta,
	complex *c, int dispc, int ldc) {
	
	cblas_zgemm (order, transA, transB, m, n, k, alpha, a+dispa, lda, b+dispb, ldb,
		beta, c+dispc, ldc);
	return 0;
}

int my_cblas_sgemm (int order, int transA, int transB, int m, int n, int k, float alpha,
	float *a, int dispa, int lda, float *b, int dispb, int ldb, float beta,
	float *c, int dispc, int ldc) {
	
	cblas_sgemm (order, transA, transB, m, n, k, alpha, a+dispa, lda, b+dispb, ldb,
		beta, c+dispc, ldc);
	return 0;
}

int my_cblas_cgemm (int order, int transA, int transB, int m, int n, int k, scomplex *alpha,
	scomplex *a, int dispa, int lda, scomplex *b, int dispb, int ldb, scomplex *beta,
	scomplex *c, int dispc, int ldc) {
	
	cblas_cgemm (order, transA, transB, m, n, k, alpha, a+dispa, lda, b+dispb, ldb,
		beta, c+dispc, ldc);
	return 0;
}




int dgeqlf_ (long int *m, long int *n, double a[], long int *lda,
	      double tau[], double work[], long int *lwork, long int *info);

int my_dgeqlf (int m, int n, double *a, int dispa, int lda, double *tau,
				double *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = -1;
	dgeqlf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}

int dgelqf_ (long int *m, long int *n, double a[], long int *lda,
	      double tau[], double work[], long int *lwork, long int *info);

int my_dgelqf (int m, int n, double *a, int dispa, int lda, double *tau,
				double *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = -1;
	dgelqf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}


int sgeqlf_ (long int *m, long int *n, float a[], long int *lda,
	      float tau[], float work[], long int *lwork, long int *info);

int my_sgeqlf (int m, int n, float *a, int dispa, int lda, float *tau,
				float *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = -1;
	sgeqlf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}

int sgelqf_ (long int *m, long int *n, float a[], long int *lda,
	      float tau[], float work[], long int *lwork, long int *info);

int my_sgelqf (int m, int n, float *a, int dispa, int lda, float *tau,
				float *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = -1;
	sgelqf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}




/* lda, lwork are for complex a and work */
int zgelqf_ (long int *m, long int *n, complex *a, long int *lda,
	      complex *tau, complex *work, long int *lwork, long int *info);

/* dormlq -> zunmlq
a, c, tau, work are complex arrays */
int zunmlq_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      complex *a, long int *lda, complex *tau, complex *c, long int *ldc,
	      complex *work, long int *lwork, long int *info);

/* a, tau, work are complex arrays */
int zgeqlf_ (long int *m, long int *n, complex *a, long int *lda,
	      complex *tau, complex *work, long int *lwork, long int *info);

/* dormql -> zunmql
a, c, tau and work are complex arrays. */
int zunmql_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      complex *a, long int *lda, complex *tau, complex *c, long int *ldc,
	      complex *work, long int *lwork, long int *info);


int my_zgeqlf (int m, int n, complex *a, int dispa, int lda, complex *tau,
				complex *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = 0;
	zgeqlf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}

int my_zgelqf (int m, int n, complex *a, int dispa, int lda, complex *tau,
				complex *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = 0;
	zgelqf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}

int my_zunmql (int side, int trans, int m, int n, int k, complex *a, int dispa, int lda,
	      complex *tau, complex *c, int dispc, int ldc, complex *work, int lwork) {
	long int mm = m, nn = n, kk = k, llda = lda, lldc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N', 'C'};
	zunmql_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &llda,
			tau, c+dispc, &lldc, work, &llwork, &info);
	return ((int) info);
}

int my_zunmlq (int side, int trans, int m, int n, int k, complex *a, int dispa, int lda,
	      complex *tau, complex *c, int dispc, int ldc, complex *work, int lwork) {
	long int mm = m, nn = n, kk = k, llda = lda, lldc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N', 'C'};
	zunmlq_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &llda,
			tau, c+dispc, &lldc, work, &llwork, &info);
	return ((int) info);
}


/* lda, lwork are for complex a and work */
int cgelqf_ (long int *m, long int *n, scomplex *a, long int *lda,
	      scomplex *tau, scomplex *work, long int *lwork, long int *info);

/* dormlq -> zunmlq
a, c, tau, work are complex arrays */
int cunmlq_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      scomplex *a, long int *lda, scomplex *tau, scomplex *c, long int *ldc,
	      scomplex *work, long int *lwork, long int *info);

/* a, tau, work are complex arrays */
int cgeqlf_ (long int *m, long int *n, scomplex *a, long int *lda,
	      scomplex *tau, scomplex *work, long int *lwork, long int *info);

/* dormql -> zunmql
a, c, tau and work are complex arrays. */
int cunmql_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      scomplex *a, long int *lda, scomplex *tau, scomplex *c, long int *ldc,
	      scomplex *work, long int *lwork, long int *info);


int my_cgeqlf (int m, int n, scomplex *a, int dispa, int lda, scomplex *tau,
				scomplex *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = 0;
	cgeqlf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}

int my_cgelqf (int m, int n, scomplex *a, int dispa, int lda, scomplex *tau,
				scomplex *work, int lwork) {
	long int mm = m, nn = n, ldaa = lda, llwork = lwork, info = 0;
	cgelqf_ (&mm, &nn, a+dispa, &ldaa, tau, work, &llwork, &info);
	return ((int) info);
}

int my_cunmql (int side, int trans, int m, int n, int k, scomplex *a, int dispa, int lda,
	      scomplex *tau, scomplex *c, int dispc, int ldc, scomplex *work, int lwork) {
	long int mm = m, nn = n, kk = k, llda = lda, lldc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N', 'C'};
	cunmql_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &llda,
			tau, c+dispc, &lldc, work, &llwork, &info);
	return ((int) info);
}

int my_cunmlq (int side, int trans, int m, int n, int k, scomplex *a, int dispa, int lda,
	      scomplex *tau, scomplex *c, int dispc, int ldc, scomplex *work, int lwork) {
	long int mm = m, nn = n, kk = k, llda = lda, lldc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N', 'C'};
	cunmlq_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &llda,
			tau, c+dispc, &lldc, work, &llwork, &info);
	return ((int) info);
}






int dormql_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      double a[], long int *lda, double tau[], double c[], long int *ldc,
	      double work[], long int *lwork, long int *info);

int my_dormql (int side, int trans, int m, int n, int k,
	      double a[], int dispa, int lda, double tau[], double c[], int dispc,
	      int ldc, double work[], int lwork) {
	long int mm = m, nn = n, kk = k, ldaa = lda, ldcc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N','T'};
	dormql_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &ldaa, tau, c+dispc,
			&ldcc, work, &llwork, &info);
	return ((int) info);
}

int dormlq_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      double a[], long int *lda, double tau[], double c[], long int *ldc,
	      double work[], long int *lwork, long int *info);

int my_dormlq (int side, int trans, int m, int n, int k,
	      double a[], int dispa, int lda, double tau[], double c[], int dispc,
	      int ldc, double work[], int lwork) {
	long int mm = m, nn = n, kk = k, ldaa = lda, ldcc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N','T'};
	dormlq_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &ldaa, tau, c+dispc,
			&ldcc, work, &llwork, &info);
	return ((int) info);
}

int sormql_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      float a[], long int *lda, float tau[], float c[], long int *ldc,
	      float work[], long int *lwork, long int *info);

int my_sormql (int side, int trans, int m, int n, int k,
	      float a[], int dispa, int lda, float tau[], float c[], int dispc,
	      int ldc, float work[], int lwork) {
	long int mm = m, nn = n, kk = k, ldaa = lda, ldcc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N','T'};
	sormql_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &ldaa, tau, c+dispc,
			&ldcc, work, &llwork, &info);
	return ((int) info);
}

int sormlq_ (char *side, char *trans, long int *m, long int *n, long int *k,
	      float a[], long int *lda, float tau[], float c[], long int *ldc,
	      float work[], long int *lwork, long int *info);

int my_sormlq (int side, int trans, int m, int n, int k,
	      float a[], int dispa, int lda, float tau[], float c[], int dispc,
	      int ldc, float work[], int lwork) {
	long int mm = m, nn = n, kk = k, ldaa = lda, ldcc = ldc, llwork = lwork, info = -1;
	char sides[] = {'L', 'R'};
	char transps[] = {'N','T'};
	sormlq_ (sides+side, transps+trans, &mm, &nn, &kk, a+dispa, &ldaa, tau, c+dispc,
			&ldcc, work, &llwork, &info);
	return ((int) info);
}





int dlacpy_(char *uplo, long int *m, long int *n, double a[], long int *lda,
	     double b[], long int *ldb);

int my_dlacpy (int uplo, int m, int n, double a[], int dispa, int lda, double b[],
				int dispb, int ldb) {
	long int mm = m, nn = n, llda = lda, lldb = ldb;
	char uplos[] = {'F', 'L', 'U'};
	dlacpy_ (uplos+uplo, &mm, &nn, a+dispa, &llda, b+dispb, &lldb);
	return 0;
}

/* lda and ldb are for complex arrays a and b */
int zlacpy_(char *uplo, long int *m, long int *n,
	     complex *a, long int *lda, complex *b, long int *ldb);

int my_zlacpy (int uplo, int m, int n, complex *a, int dispa, int lda, complex *b,
				int dispb, int ldb) {
	long int mm = m, nn = n, llda = lda, lldb = ldb;
	char uplos[] = {'F', 'L', 'U'};
	zlacpy_ (uplos+uplo, &mm, &nn, a+dispa, &llda, b+dispb, &lldb);
	return 0;
}

int slacpy_(char *uplo, long int *m, long int *n, float a[], long int *lda,
	     float b[], long int *ldb);

int my_slacpy (int uplo, int m, int n, float a[], int dispa, int lda, float b[],
				int dispb, int ldb) {
	long int mm = m, nn = n, llda = lda, lldb = ldb;
	char uplos[] = {'F', 'L', 'U'};
	slacpy_ (uplos+uplo, &mm, &nn, a+dispa, &llda, b+dispb, &lldb);
	return 0;
}

/* lda and ldb are for complex arrays a and b */
int clacpy_(char *uplo, long int *m, long int *n,
	     scomplex *a, long int *lda, scomplex *b, long int *ldb);

int my_clacpy (int uplo, int m, int n, scomplex *a, int dispa, int lda, scomplex *b,
				int dispb, int ldb) {
	long int mm = m, nn = n, llda = lda, lldb = ldb;
	char uplos[] = {'F', 'L', 'U'};
	clacpy_ (uplos+uplo, &mm, &nn, a+dispa, &llda, b+dispb, &lldb);
	return 0;
}





int my_dsub (int m, int n, double *a, int dispa, int lda, double *b, int dispb, int ldb,
			double *c, int dispc, int ldc) {
	int i, j;
	double *aa = a + dispa;
	double *bb = b + dispb;
	double *cc = c + dispc;
	for (j = 0; j < n; j++) {
		for (i = 0; i < m; i++) {
			cc[ldc * j + i] = aa[lda*j + i] - bb[ldb*j + i];
		};
	};
	return 0;		
}

int my_zsub (int m, int n, complex *a, int dispa, int lda,
			complex *b, int dispb, int ldb,
			complex *c, int dispc, int ldc) {
	int i, j;
	complex *aa = a + dispa;
	complex *bb = b + dispb;
	complex *cc = c + dispc;
	for (j = 0; j < n; j++) {
		for (i = 0; i < m; i++) {
			cc[ldc * j + i].re = aa[lda*j + i].re - bb[ldb*j + i].re;
			cc[ldc * j + i].im = aa[lda*j + i].im - bb[ldb*j + i].im;
		};
	};
	return 0;		
}

int my_ssub (int m, int n, float *a, int dispa, int lda, float *b, int dispb, int ldb,
			float *c, int dispc, int ldc) {
	int i, j;
	float *aa = a + dispa;
	float *bb = b + dispb;
	float *cc = c + dispc;
	for (j = 0; j < n; j++) {
		for (i = 0; i < m; i++) {
			cc[ldc * j + i] = aa[lda*j + i] - bb[ldb*j + i];
		};
	};
	return 0;		
}

int my_csub (int m, int n, scomplex *a, int dispa, int lda,
			scomplex *b, int dispb, int ldb,
			scomplex *c, int dispc, int ldc) {
	int i, j;
	scomplex *aa = a + dispa;
	scomplex *bb = b + dispb;
	scomplex *cc = c + dispc;
	for (j = 0; j < n; j++) {
		for (i = 0; i < m; i++) {
			cc[ldc * j + i].re = aa[lda*j + i].re - bb[ldb*j + i].re;
			cc[ldc * j + i].im = aa[lda*j + i].im - bb[ldb*j + i].im;
		};
	};
	return 0;		
}


extern int dgels_ (char *trans, long int *m, long int *n, long int *nrhs, double a[],
	     long int *lda, double b[], long int *ldb, double work[], long int *lwork, long int *info);

int my_dgels (int trans, int m, int n, int nrhs, double *a, int dispa, int lda,
				double *b, int dispb, int ldb, double *work, int lwork) {
	long int mm = m, nn = n, nrhss = nrhs, llwork = lwork, llda = lda, lldb = ldb,
			info = -1;
	char transps[] = {'N', 'T'};
	dgels_ (transps+trans, &mm, &nn, &nrhss, a+dispa, &llda, b+dispb, &lldb, work,
			&llwork, &info);
	return ((int) info);
}

extern int zgels_ (char *trans, long int *m, long int *n, long int *nrhs, complex *a,
	     long int *lda, complex *b, long int *ldb, complex *work,long int *lwork, long int *info);

int my_zgels (int trans, int m, int n, int nrhs, complex *a, int dispa, int lda,
				complex *b, int dispb, int ldb, complex *work, int lwork) {
	long int mm = m, nn = n, nrhss = nrhs, llwork = lwork, llda = lda, lldb = ldb,
			info = -1;
	char transps[] = {'N', 'C'};
	zgels_ (transps+trans, &mm, &nn, &nrhss, a+dispa, &llda, b+dispb, &lldb, work,
			&llwork, &info);
	return ((int) info);
}

extern int sgels_ (char *trans, long int *m, long int *n, long int *nrhs, float a[],
	     long int *lda, float b[], long int *ldb, float work[], long int *lwork, long int *info);

int my_sgels (int trans, int m, int n, int nrhs, float *a, int dispa, int lda,
				float *b, int dispb, int ldb, float *work, int lwork) {
	long int mm = m, nn = n, nrhss = nrhs, llwork = lwork, llda = lda, lldb = ldb,
			info = -1;
	char transps[] = {'N', 'T'};
	sgels_ (transps+trans, &mm, &nn, &nrhss, a+dispa, &llda, b+dispb, &lldb, work,
			&llwork, &info);
	return ((int) info);
}

extern int cgels_ (char *trans, long int *m, long int *n, long int *nrhs, scomplex *a,
	     long int *lda, scomplex *b, long int *ldb, scomplex *work,long int *lwork, long int *info);

int my_cgels (int trans, int m, int n, int nrhs, scomplex *a, int dispa, int lda,
				scomplex *b, int dispb, int ldb, scomplex *work, int lwork) {
	long int mm = m, nn = n, nrhss = nrhs, llwork = lwork, llda = lda, lldb = ldb,
			info = -1;
	char transps[] = {'N', 'C'};
	cgels_ (transps+trans, &mm, &nn, &nrhss, a+dispa, &llda, b+dispb, &lldb, work,
			&llwork, &info);
	return ((int) info);
}




extern int dtrtrs_ (char *uplo, char *trans, char *diag, long int *n, long int *nrhs,
	      double a[], long int *lda, double b[], long int *ldb, long int *info);

/*  a and b are complex arrays. */
extern int ztrtrs_ (char *uplo, char *trans, char *diag, long int *n, long int *nrhs,
	      complex a[], long int *lda, complex b[], long int *ldb, long int *info);

int my_dtrtrs (int uplo, int trans, int diag, int n, int nrhs,
	      double a[], int dispa, int lda, double b[], int dispb, int ldb) {
	long int nn = n, nnrhs = nrhs, llda = lda, lldb = ldb, info = -1;
	char uplos[] = {'L', 'U'};
	char transps[] = {'N', 'T'};
	char diags[] = {'N', 'U'};
	dtrtrs_ (uplos+uplo, transps+trans, diags+diag, &nn, &nnrhs, a+dispa, &llda,
			b+dispb, &lldb, &info);
	return ((int) info);
}

int my_ztrtrs (int uplo, int trans, int diag, int n, int nrhs,
	      complex a[], int dispa, int lda, complex b[], int dispb, int ldb) {
	long int nn = n, nnrhs = nrhs, llda = lda, lldb = ldb, info = -1;
	char uplos[] = {'L', 'U'};
	char transps[] = {'N', 'C'};
	char diags[] = {'N', 'U'};
	ztrtrs_ (uplos+uplo, transps+trans, diags+diag, &nn, &nnrhs, a+dispa, &llda,
			b+dispb, &lldb, &info);
	return ((int) info);
}

extern int strtrs_ (char *uplo, char *trans, char *diag, long int *n, long int *nrhs,
	      float a[], long int *lda, float b[], long int *ldb, long int *info);

/*  a and b are complex arrays. */
extern int ctrtrs_ (char *uplo, char *trans, char *diag, long int *n, long int *nrhs,
	      scomplex a[], long int *lda, scomplex b[], long int *ldb, long int *info);

int my_strtrs (int uplo, int trans, int diag, int n, int nrhs,
	      float a[], int dispa, int lda, float b[], int dispb, int ldb) {
	long int nn = n, nnrhs = nrhs, llda = lda, lldb = ldb, info = -1;
	char uplos[] = {'L', 'U'};
	char transps[] = {'N', 'T'};
	char diags[] = {'N', 'U'};
	strtrs_ (uplos+uplo, transps+trans, diags+diag, &nn, &nnrhs, a+dispa, &llda,
			b+dispb, &lldb, &info);
	return ((int) info);
}

int my_ctrtrs (int uplo, int trans, int diag, int n, int nrhs,
	      scomplex a[], int dispa, int lda, scomplex b[], int dispb, int ldb) {
	long int nn = n, nnrhs = nrhs, llda = lda, lldb = ldb, info = -1;
	char uplos[] = {'L', 'U'};
	char transps[] = {'N', 'C'};
	char diags[] = {'N', 'U'};
	ctrtrs_ (uplos+uplo, transps+trans, diags+diag, &nn, &nnrhs, a+dispa, &llda,
			b+dispb, &lldb, &info);
	return ((int) info);
}



extern int dgesvd_ (char *jobu, char *jobvt, long int *m, long int *n, double a[],
	      long int *lda, double s[], double u[], long int *ldu, double vt[],
	      long int *ldvt, double work[], long int *lwork, long int *info);

int my_dgesvd (int jobu, int jobvt, int m, int n, double *a, int dispa, int lda,
				double *s, double *u, int dispu, int ldu, double *vt,
				int dispvt, int ldvt, double *work, int lwork) {
	long int mm = m, nn = n, llda = lda, lldu = ldu, lldvt = ldvt, llwork = lwork, info = -1;
	char jobs[] = {'A', 'S', 'O', 'N'};
	dgesvd_ (jobs+jobu, jobs+jobvt, &mm, &nn, a+dispa, &llda, s, u+dispu, &lldu,
			vt+dispvt, &lldvt, work, &llwork, &info);
	return ((int) info);
}

/* a, u, vt, work are complex arrays.
   double rwork[5*min(M,N)] */
extern int zgesvd_ (char *jobu, char *jobvt, long int *m, long int *n,
			complex a[], long int *lda, double s[], complex u[], long int *ldu,
	      	complex vt[], long int *ldvt, complex work[], long int *lwork,
			double rwork[], long int *info);

int my_zgesvd (int jobu, int jobvt, int m, int n, complex *a, int dispa, int lda,
				double *s, complex *u, int dispu, int ldu, complex *vt,
				int dispvt, int ldvt, complex *work, int lwork, double *rwork) {
	long int mm = m, nn = n, llda = lda, lldu = ldu, lldvt = ldvt, llwork = lwork, info = -1;
	char jobs[] = {'A', 'S', 'O', 'N'};
	zgesvd_ (jobs+jobu, jobs+jobvt, &mm, &nn, a+dispa, &llda, s, u+dispu, &lldu,
			vt+dispvt, &lldvt, work, &llwork, rwork, &info);
	return ((int) info);
}

extern int sgesvd_ (char *jobu, char *jobvt, long int *m, long int *n, float a[],
	      long int *lda, float s[], float u[], long int *ldu, float vt[],
	      long int *ldvt, float work[], long int *lwork, long int *info);

int my_sgesvd (int jobu, int jobvt, int m, int n, float *a, int dispa, int lda,
				float *s, float *u, int dispu, int ldu, float *vt,
				int dispvt, int ldvt, float *work, int lwork) {
	long int mm = m, nn = n, llda = lda, lldu = ldu, lldvt = ldvt, llwork = lwork, info = -1;
	char jobs[] = {'A', 'S', 'O', 'N'};
	sgesvd_ (jobs+jobu, jobs+jobvt, &mm, &nn, a+dispa, &llda, s, u+dispu, &lldu,
			vt+dispvt, &lldvt, work, &llwork, &info);
	return ((int) info);
}

/* a, u, vt, work are complex arrays.
   double rwork[5*min(M,N)] */
extern int cgesvd_ (char *jobu, char *jobvt, long int *m, long int *n,
			scomplex a[], long int *lda, float s[], scomplex u[], long int *ldu,
	      	scomplex vt[], long int *ldvt, scomplex work[], long int *lwork,
			float rwork[], long int *info);

int my_cgesvd (int jobu, int jobvt, int m, int n, scomplex *a, int dispa, int lda,
				float *s, scomplex *u, int dispu, int ldu, scomplex *vt,
				int dispvt, int ldvt, scomplex *work, int lwork, float *rwork) {
	long int mm = m, nn = n, llda = lda, lldu = ldu, lldvt = ldvt, llwork = lwork, info = -1;
	char jobs[] = {'A', 'S', 'O', 'N'};
	cgesvd_ (jobs+jobu, jobs+jobvt, &mm, &nn, a+dispa, &llda, s, u+dispu, &lldu,
			vt+dispvt, &lldvt, work, &llwork, rwork, &info);
	return ((int) info);
}



int my_cblas_dswap (int N, double *X, int dispX, int incX, double *Y,
		    int dispY, int incY) {
  cblas_dswap (N, X+dispX, incX, Y+dispY, incY);
  return 0;
}

int my_cblas_zswap (int N, complex *X, int dispX, int incX, complex *Y,
		    int dispY, int incY) {
  cblas_zswap (N, X+dispX, incX, Y+dispY, incY);
  return 0;
}

extern int dgetrf_ (long int *m, long int *n, double a[], long int *lda,
		long int ipiv[], long int *info);

int my_dgetrf (int m, int n, double *a, int dispa, int lda, int *ipiv) {
	long int mm = m, nn = n, llda = lda, info;
	long int *iipiv;
	int i, k = m < n ? m : n;
	iipiv = (long int *) malloc (sizeof (long int) * k);
	for (i=0; i< k; i++) {
		iipiv[i] = -1;
	};
	dgetrf_ (&mm, &nn, a+dispa, &llda, iipiv, &info);
	for (i=0; i<k; i++) {
		ipiv[i] = (int) iipiv[i];
	};
	free ((void *) iipiv);
	return((int) info);
}

extern int zgetrf_ (long int *m, long int *n, complex a[], long int *lda,
		long int ipiv[], long int *info);

int my_zgetrf (int m, int n, complex *a, int dispa, int lda, int *ipiv) {
	long int mm = m, nn = n, llda = lda, info;
	long int *iipiv;
	int i, k = m < n ? m : n;
	iipiv = (long int *) malloc (sizeof (long int) * k);
	for (i=0; i< k; i++) {
		iipiv[i] = -1;
	};
	zgetrf_ (&mm, &nn, a+dispa, &llda, iipiv, &info);
	for (i=0; i<k; i++) {
		ipiv[i] = (int) iipiv[i];
	};
	free ((void *) iipiv);
	return((int) info);
}

/* LWORK >= 3*N+1 */
extern int dgeqp3_ (long int *m, long int *n, double a[], long int *lda,
		long int jpvt[], double tau[], double work[], long int *lwork, long int *info);

int my_dgeqp3 (int m, int n, double *a, int dispa, int lda, int *jpvt,
		double *tau, double *work, int lwork) {
	long int mm = m, nn = n, llda = lda, llwork = lwork, info;
	long int *jjpvt;
	int i;
	jjpvt = (long int *) malloc(sizeof(long int) * n);
	for (i=0; i<n; i++) {
		jjpvt[i] = (long int) jpvt[i];
	};
	dgeqp3_ (&mm, &nn, a+dispa, &llda, jjpvt, tau, work, &llwork, &info);
	for (i=0; i<n; i++) {
		jpvt[i] = (int) jjpvt[i];
	};
	free((void *) jjpvt);
	return ((int) info);
}

extern int zgeqp3_ (long int *m, long int *n, complex a[], long int *lda,
		long int jpvt[], complex tau[], complex work[], long int *lwork,
		double rwork[], long int *info);

int my_zgeqp3 (int m, int n, complex *a, int dispa, int lda, int *jpvt,
		complex *tau, complex *work, int lwork, double *rwork) {
	long int mm = m, nn = n, llda = lda, llwork = lwork, info;
	long int *jjpvt;
	int i;
	jjpvt = (long int *) malloc(sizeof(long int) * n);
	for (i=0; i<n; i++) {
		jjpvt[i] = (long int) jpvt[i];
	};
	zgeqp3_ (&mm, &nn, a+dispa, &llda, jjpvt, tau, work, &llwork, rwork, &info);
	for (i=0; i<n; i++) {
		jpvt[i] = (int) jjpvt[i];
	};
	free((void *) jjpvt);
	return ((int) info);
}


double *ddup (double a[]) {
	return a;
}


complex *zdup (complex a[]) {
	return a;
}


