/*
	File: complex.icl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Complex numbers
	See also: CamlFloat
	Date: 6.17.2003
	
	$Id$
*/


implementation module complex

import StdEnv

// Tuples cannot be unboxed in arrays. So use records. Cannot make this generic.
:: Complex = {re :: !Real, im :: !Real}

realPart :: !Complex -> Real
realPart z = z.re

imagPart :: !Complex -> Real
imagPart z = z.im

instance fromInt Complex
where
	fromInt :: Int -> Complex
	fromInt n = {re = fromInt n, im = fromInt 0}

(+:) infix 6 :: !Real !Real -> Complex
(+:) x y = {re=x, im=y}

instance zero Complex
where
	zero = 0.0 +: 0.0

instance one Complex
where
	one = 1.0 +: 0.0

instance + Complex
where
	(+) z1 z2 = {re = z1.re + z2.re, im = z1.im + z2.im}

instance - Complex
where
	(-) z1 z2 = {re = z1.re - z2.re, im = z1.im - z2.im}


instance toString Complex
where
	toString z = if (z.im > 0.0)
					(toString z.re +++ "+i" +++ toString z.im)
					(toString z.re +++ "-i" +++ toString (abs z.im))

conjg :: !Complex -> Complex
conjg z = {re = z.re, im = ~z.im}

instance * Complex
where
	(*) z1 z2 = { re = z1.re * z2.re - z1.im * z2.im
				, im = z1.re * z2.im + z1.im * z2.re
				}
