/*
	File: context.icl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Contextual zippers (sparse arrays optimized for sequential access)
	See also: CamlFloat
	Date: 10.03.2003
	
	$Id$
*/

implementation module context

import StdEnv, StdOverloadedList

:: Context e = { cHd :: ![!(Int,e)!], cTl :: ![!(Int,e)!] }
:: Option e = None | Some !e

cempty :: Context e
cempty = { cHd = [!!], cTl = [!!] }


cget :: !(Context e) !Int -> (Option e, Context e)
cget c=:{ cHd = [!!], cTl = [!!]} i = (None, c)
cget c=:{ cHd = [!(j,ej) : hds!] } i
	| j == i = (Some ej, { cHd = hds, cTl = [!(j,ej) : c.cTl!] })
	| j < i = cget { cHd = hds, cTl = [! (j,ej) : c.cTl !] } i
	| j > i = tget c i
cget c=:{ cTl = [!(j,ej) : tls!] } i
	| j == i = (Some ej, { cHd = [! (j,ej) : c.cHd !], cTl = tls } )
	| j > i = tget { cHd = [! (j,ej) : c.cHd !], cTl = tls } i
	| j < i = (None, c)

tget c=:{ cTl = [!!] } i = (None, c)
tget c=:{ cTl = [! (j,ej) : tls !] } i
	| j == i = (Some ej, { cHd = [! (j,ej) : c.cHd !], cTl = tls } )
	| j > i = tget { cHd = [! (j,ej) : c.cHd !], cTl = tls } i
	| j < i = (None, c)


cset :: !(Context e) !Int !(Option e) -> Context e
cset { cHd = [!!], cTl = [!!]} i (Some e) = { cHd = [!(i,e)!], cTl = [!!] }
cset c=:{ cHd = [!!], cTl = [!!]} i None = c
cset c=:{ cHd = [!(j,ej) : hds!] } i v
	| j == i = case v of
				Some e = { cHd = hds, cTl = [!(j,e) : c.cTl!] }
				None = { cHd = hds, cTl = c.cTl }
	| j < i = cset { cHd = hds, cTl = [! (j,ej) : c.cTl !] } i v
	| j > i = tset c i v
cset c=:{ cTl = [!(j,ej) : tls!] } i v
	| j == i = case v of
				 Some e = { cHd = [! (j,e) : c.cHd !], cTl = tls }
				 None = { cHd = c.cHd, cTl = tls }
	| j > i = tset { cHd = [! (j,ej) : c.cHd !], cTl = tls } i v
	| j < i = case v of
				Some e = { cHd = c.cHd, cTl = [! (i,e) : c.cTl !] }
				None = c

tset c=:{ cTl = [!!] } i (Some e) = { cHd = c.cHd, cTl = [!(i,e)!] }
tset c=:{ cTl = [!!] } i None = c
tset c=:{ cTl = [! (j,ej) : tls !] } i v
	| j == i = case v of
				Some e = { cHd = [! (j,e) : c.cHd !], cTl = tls }
				None = { cHd = c.cHd, cTl = tls }
	| j > i = tset { cHd = [! (j,ej) : c.cHd !], cTl = tls } i v
	| j < i = case v of
				Some e = { cHd = c.cHd, cTl = [! (i,e) : c.cTl !] }
				None = c


choose :: !(Context e) -> Option (Int, e)
choose { cHd = [!!], cTl = [!!] } = None
choose { cHd = [! v : _ !] } = Some v
choose { cTl = [! v : _ !] } = Some v

first :: !(Context e) -> (Option (Int, e), Context e)
first c=:{ cHd = [!!], cTl = [!!] } = (None, c)
first c = let [! v : hds !] = Reverse c.cTl ++| c.cHd
			  d = { cHd = hds, cTl = [! v !] }
		  in
			(Some v, d)

kadesi :: !(Context e) -> (Option (Int, e), Context e)
kadesi c=:{ cHd = [!!], cTl = [!!] } = (None, c)
kadesi c = let [! v : tls !] = Reverse c.cHd ++| c.cTl
			   d = { cHd = [! v !], cTl = tls }
		   in
			(Some v, d)

next :: !(Context e) -> (Option (Int, e), Context e)
next c=:{ cHd = [!!] } = (None, c)
next c=:{ cHd = [! v : hds !] } = (Some v, { cHd = hds, cTl = [! v : c.cTl !] })

prev :: !(Context e) -> (Option (Int, e), Context e)
prev c=:{ cTl = [!!] } = (None, c)
prev c=:{ cTl = [! v : tls !] } = (Some v, { cHd = [! v : c.cHd !], cTl = tls })

toList :: !(Context e) -> [!(Int, e)!]
toList c = Reverse c.cTl ++| c.cHd

fromList :: ![!(Int,e)!] -> Context e
fromList es = bld es cempty
where
	bld [!!] c = c
	bld [!(i,v) : es!] c = bld es (cset c i (Some v))

test1 =
	let c = cset (cset (cset (cset (cset (cset (cset cempty 1 (Some 'C'))
						4 (Some 'A')) 6 (Some 'T')) 2 (Some 'H')) 6 None) 7
							(Some 'T')) 4 (Some 'I')
		(v, d) = first c
		(v2, d2) = next d
		(v3, d3) = next d2
	in
		(v, v2, v3)

test2 =
	let c = cset (cset (cset (cset (cset (cset (cset cempty 1 (Some 'C'))
						4 (Some 'A')) 6 (Some 'T')) 2 (Some 'H')) 6 None) 7
							(Some 'T')) 4 (Some 'I')
	in
		toList c

test3 = fromList [! (4,'T'), (10, 'R'), (3, 'A'), (2, 'C'), (12, 'H') !]

test4 =
	let c = cset
				(cset
					(cset
						(cset
							(cset
								(cset
									(cset cempty 1 (Some 'C'))
									4 (Some 'A'))
								6 (Some 'T'))
							2 (Some 'H'))
						6 None)
					7 (Some 'T'))
				4 (Some 'I')
		(v, d) = kadesi c
		(v2, d2) = prev d
		(v3, _) = prev d2
	in
		(v3, v2, v)


Start = test4
