/*
	File: sparse.dcl
	Project: CleanFloat
	Author: Shivkumar Chandrasekaran
	Email: shiv@ece.ucsb.edu
	Purpose: Block sparse matrices
	See also: CamlFloat
	Date: 10.04.2003
	
	$Id$
*/

definition module sparse

import StdEnv, blop, context

:: Sparse e = Sparse !.{Context (Matrix e)} !.{Context (Matrix e)}
:: SparseTransform e

// skeleton m n creates an empty sparse matrix with m row partitions and n
// column partitions.
skeleton :: !Int !Int -> *(Sparse e)

instance noOfRows  Sparse e | noOfRows Matrix e
instance noOfCols  Sparse e | noOfCols Matrix e
instance noOfRows` Sparse e | noOfRows Matrix e
instance noOfCols` Sparse e | noOfCols Matrix e

getBlock  :: !(Sparse e)  !Int !Int -> Option (Matrix e)
getBlock` :: !*(Sparse e) !Int !Int -> (Option (Matrix e), *(Sparse e))

setBlock :: !*(Sparse e) !Int !Int !(Option (Matrix e)) -> *(Sparse e)

rowPartitions  :: !(Sparse e)  -> {#Int}                | noOfRows Matrix e
rowPartitions` :: !*(Sparse e) -> ({#Int}, *(Sparse e)) | noOfRows Matrix e
colPartitions  :: !(Sparse e)  -> {#Int}                | noOfCols Matrix e
colPartitions` :: !*(Sparse e) -> ({#Int}, *(Sparse e)) | noOfCols Matrix e

instance fastMul Sparse e | zero, fun2mat, MatrixAccess, MatrixArithU e

// ql factorize a sparse matrix. The spine of the sparse matrix is re-used
// to hold the l factor. The blocks of the sparse matrix are not destroyed.
// The q factor is correct only for square matrices with square diagonal
// blocks.
qlFactor :: !*(Sparse e) -> (SparseTransform e, Sparse e) | MatrixAccess
														  , MatrixArithU e

// Apply a sparse transform from the left
transfL :: !(SparseTransform e) !*(Matrix e) -> *(Matrix e) | MatrixAccess
															, MatrixArithU e

// Apply the conjugate transpose of a sparse transform from the left
transfLT :: !(SparseTransform e) !*(Matrix e) -> *(Matrix e) | MatrixAccess
															 , MatrixArithU e

// fSubs lsp a -> lsp^{-1} a, where lsp is sparse lower-triangular.
fSubs :: !(Sparse e) *(Matrix e) -> *(Matrix e) | MatrixAccess, MatrixArithU e
